/*
Title- Exploring Correlation for Indirect Branch Prediction
Authors- Nikunj Bhansali - NC State University- nsbhansa@ncsu.edu
	 Chintan Panirwala - NC State University-cdpanirw@ncsu.edu
         Huiyang Zhou- NC State University- hzhou@ncsu.edu

Date-  May 2011

Code has been inspired by LTAGE predictor by Dr. Andre Seznec.
*/
// Author: Hongliang Gao;   Created: Jan 27 2011
// Description: sample predictors for cbp3.
#include <iostream>
#include<stdlib.h>
#include <fstream>
#include <assert.h>
#include <stdio.h>
#include <cassert>
#include <string.h>
#include <inttypes.h>
using namespace std;
#include "cbp3_def.h"
#include "cbp3_framework.h"
#include "predictor.h"

// this file includes two sample predictors
// one is a 64 KB gshare conditional predictor
// the other is a 64 KB indirect predictor indexed by (pc ^ history)

// rewind_marked is also used to show how to rewind the reader for multiple runs
// the predictor will use gshare in the first run
// the indirect predictor will be used in the second run

// NOTE: rewind_marked is only provided to help tuning work. the final
// submitted code should only include one run.

my_predictor *p; 
// count number of runs
uint32_t runs;
unsigned int global_value_src,global_value_pc;
void PredictorInit() {
  runs = 0;
  p = new my_predictor();
  assert(1);
}

void PredictorReset() {
    // this function is called before EVERY run
    // it is used to reset predictors and change configurations
}

//branch_predictor *p= new my_predictor();
void PredictorRunACycle() {
    // get info about what uops are processed at each pipeline stage
    const cbp3_cycle_activity_t *cycle_info = get_cycle_info();

    // make prediction at fetch stage
    for (int i = 0; i < cycle_info->num_fetch; i++) {
        uint32_t fe_ptr = cycle_info->fetch_q[i];
        const cbp3_uop_dynamic_t *uop = &fetch_entry(fe_ptr)->uop;
	branch_update *u;
	if (runs == 0 && (uop->type & IS_BR_INDIRECT)) {
	  u = p-> predict_fetch(uop->pc);
	  uint32_t gpred1 = u-> target_prediction();

	  assert(report_pred(fe_ptr, false, gpred1));
	}
	//if(uop_is_branch(uop->type))
	if((uop->type & IS_BR_CONDITIONAL)|| (uop->type & IS_BR_INDIRECT))
	  p->update_hist_fetch(uop, uop->br_taken);
    }

    //exploting address-target corelation: predict at AGEN stage of LD insns
    for (uint32_t i = 0; i < cycle_info->num_agu; i ++) {

      uint32_t rob_ptr = cycle_info->agu_q[i];
      const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;
      unsigned int exec_pred=0;
      if (uop->type & IS_LOAD)
      {
        for (int tempi=4;tempi>=1;tempi--)
        {
	  if(rob_entry(rob_ptr+tempi)->valid == 0)
	    continue;
          if ((rob_entry(rob_ptr+tempi)->uop.type) & IS_BR_INDIRECT)
          {
	    const uint32_t *last_pred =&rob_entry(rob_ptr+tempi)->last_pred;
            for (int j=0;j<6;j++)
            {
              if (rob_entry(rob_ptr+tempi)->uop.srcs_static[j]==uop->dst_static && uop->dst_static != 0)
                {
                  unsigned int global_value_src1=0;
 
		  global_value_src1=(uop->vaddr ^ (uop->vaddr >> 10)) & ((1<<(10))-1);
                  if (global_value_src1 != 0)
                  {
                    exec_pred=p->pred_execute(&rob_entry(rob_ptr+tempi)->uop,global_value_src1);
                  }

                  if (exec_pred!=0 && *last_pred!=exec_pred)
                  {
		    assert(report_pred(rob_ptr+tempi, true,exec_pred ));
	          }
                }
            }
          }
        }
      }
    }

    //exploting address-target corelation: update at EXE stage of Indirect branches
    for (uint32_t i = 0; i < cycle_info->num_exe; i ++) {
      
      uint32_t rob_ptr = cycle_info->exe_q[i];
      const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;
      
       if (uop->type & IS_BR_INDIRECT)
      {
	if((rob_entry(rob_ptr)->cycle_last_pred == rob_entry(rob_ptr)->cycle_fetch && uop->br_target != rob_entry(rob_ptr)->last_pred)||(rob_entry(rob_ptr)->cycle_last_pred != rob_entry(rob_ptr)->cycle_fetch))
	{
	  for(int tempi = 1; tempi <= 4; tempi++)
	  {
	    if(rob_entry(rob_ptr-tempi)->valid == 0)
	      continue;
	    if ((rob_entry(rob_ptr-tempi)->uop.type) & IS_LOAD)
	    {
	      for (int j=0;j<6;j++)
	      {
		if (rob_entry(rob_ptr-tempi)->uop.dst_static == uop->srcs_static[j] && uop->srcs_static[j] != 0)
		{
		  global_value_src = rob_entry(rob_ptr-tempi)->uop.vaddr;
		  global_value_src = (global_value_src ^ (global_value_src >> 10)) & ((1<<(10))-1);

		  global_value_pc = uop->pc;
		 
		  p->pred_exe_update (uop,global_value_src,uop->br_target);
		}
	      }
	    }
	  }
	}
      }     
    }
        
    for (int i = 0; i < cycle_info->num_retire; i++) {
        uint32_t rob_ptr = cycle_info->retire_q[i];
        const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;
	//int n_br = 0;
	branch_update *u;
  
	if (runs == 0 && (uop->type & IS_BR_INDIRECT)) {  
	  u = p -> predict_retire(uop);
	  p->update_retire(u, uop->br_taken,uop->br_target,0); 
        }

	if((uop->type & IS_BR_CONDITIONAL)|| (uop->type & IS_BR_INDIRECT))
	  p->update_hist_retire(uop, uop->br_taken);
	
    }
    
}

void PredictorRunEnd() {
    runs ++;
    p->init();
    if (runs < 1) // set rewind_marked to indicate that we want more runs
        rewind_marked = true;
}

void PredictorExit() {

}
